//----------------------------------------------------------------------------
/// @file parallel_sort.hpp
/// @brief Parallel Sort algorithm
///
/// @author Copyright (c) 2010 Francisco José Tapia (fjtapia@gmail.com )\n
///         Distributed under the Boost Software License, Version 1.0.\n
///         ( See accompanyingfile LICENSE_1_0.txt or copy at
///           http://www.boost.org/LICENSE_1_0.txt  )
/// @version 0.1
///
/// @remarks
//-----------------------------------------------------------------------------
#ifndef __CNTREE_SORT_PARALLEL_SORT_HPP
#define __CNTREE_SORT_PARALLEL_SORT_HPP

#include <algorithm>
#include <boost/countertree/util/atomic.hpp>
#include <boost/countertree/util/stack_cnc.hpp>
#include <boost/countertree/util/sort/token.hpp>
#include <boost/countertree/util/sort/indirect_sort.hpp>
#include <vector>


namespace boost
{
namespace countertree
{
namespace util
{
namespace sort
{
using namespace boost::countertree::util;

template    < class iter_t,
              typename compare = std::less < typename iter_value<iter_t>::type>
            >
struct parallel_sort_comp
{   //------------------------- begin ----------------------
    typedef  token_level < iter_t> token_t ;
    size_type MaxPerThread = 65536;
    static const size_t Min_Parallel = 65536 ;
    compare comp;

    //----------------------- Variables --------------------------------------
    stack_cnc <token_t> ST ;
    std::atomic<difference_type > ND ;

    //----------------------- Constructor--------------------------------------
    parallel_sort_comp (iter_t first, iter_t last, NThread NT= NThread() )
                        :parallel_sort_comp ( first,last, compare(),NT){} ;

    parallel_sort_comp ( iter_t first,
                         iter_t last,
                         compare comp1,
                         NThread NT= NThread())
                        :comp(comp1)
    {   //------------------------- begin ----------------------
        difference_type N = last - first;
        assert ( N >=0);
        if ( (size_t)N < Min_Parallel or NT() == 1 )
        {   std::sort ( first, last,comp) ;
            return ;
        } ;

        //-------------------------------------------------------------------
        MaxPerThread = N / (NT() *8); //8 parts for each thread
        if ( MaxPerThread < Min_Parallel) MaxPerThread = Min_Parallel ;

        uint32_t Level = (NBits (N) - 16 ) <<1 ;
        ST.reserve ( Level * NT() );
        std::atomic_store (&ND, N);
        ST.emplace_back ( first, last, Level);
        std::vector<std::thread> VT (NT());
        for ( uint32_t i  =0  ; i < NT() ; ++i)
            VT[i] = std::thread ( &parallel_sort_comp::sort_thread, this );
        for ( uint32_t i  =0  ; i < NT() ; ++i)
            VT[i].join();
    };

    void sort_thread( )
    {   //----------------------------- begin -------------------------------
        token_t tk ;
        while ( atomic_read ( ND) != 0)
        {   //----------------- read from ST --------------------------------
            if ( not ST.pop_copy_back( tk) )
            {   std::this_thread::yield() ;
                continue ;
            };
            difference_type N = tk.last - tk.first ;
            if ( (size_type)N <= MaxPerThread or tk.level == 0)
            {   atomic_sub ( ND, N );
                std::sort ( tk.first, tk.last, comp) ;
                continue ;
            };

            //-------------------  split ------------------------------------
            typedef typename iter_value<iter_t>::type  value_t ;
            value_t val = const_cast < value_t &> (*(tk.first +(N>>1)) );

            iter_t c_first = tk.first , c_last  = tk.last-1;
            for (; comp(*c_first, val) ; ++c_first) ;
            for (; comp(val ,*c_last ) ; --c_last );
            while (not( c_first > c_last ))
            {   std::swap ( *(c_first++), *(c_last--));
                for ( ;comp(*c_first,val) ; ++c_first);
                for ( ;comp(val, *c_last) ; --c_last );
            }; // End while

            //---------------- Store the tokens in ST ------------------------
            token_t tk1 ( tk.first, c_last+1, tk.level-1) ;
            token_t tk2 ( c_first , tk.last,  tk.level-1) ;
            auto N1 = (tk1.last - tk1.first);
            auto N2 = (tk2.last - tk2.first);
            auto N3 = N - N1- N2 ;
            if ( N3 != 0) atomic_sub ( ND, N3 );
            if ( N2 > N1 )
            {   ST.push_back ( tk2);
                ST.push_back ( tk1);
            }
            else
            {   ST.push_back ( tk1);
                ST.push_back ( tk2);
            } ;
        }; // end while
    } ;
};
//------------------------------------------------------------------------------
// These functions are for to select the correct format depending of the number
// and type of the parameters
//------------------------------------------------------------------------------
template    < class iter_t,
              typename compare = std::less <typename iter_value<iter_t>::type>
            >
void parallel_sort ( iter_t first, iter_t last , NThread NT = NThread() )
{   parallel_sort_comp <iter_t> ( first, last, NT);
};

template    < class iter_t,
              typename compare = std::less <typename iter_value<iter_t>::type>
            >
void parallel_sort ( iter_t first, iter_t last,
                     compare comp1 , NThread NT = NThread() )
{   //------------------------------- begin ---------------------------------
    parallel_sort_comp<iter_t,compare> ( first, last,comp1,NT);
};

template < class iter_t,
          typename compare = std::less < typename iter_value<iter_t>::type >
        >
void indirect_parallel_sort (iter_t first, iter_t last, NThread NT=NThread() )
{  	//-------------------------------- begin ----------------------------------
    parallel_indirect_sort_method<parallel_sort_comp, iter_t,compare>
                                                  ( first , last, compare() );
};

//
//****************************************************************************
};//    End namespace sort
};//    End namespace util
};//    End namespace countertree
};//    End namespace boost
//****************************************************************************
//
#endif
