//----------------------------------------------------------------------------
/// @file merge_sort.hpp
/// @brief Quick Sort algorithm
///
/// @author Copyright (c) 2010 Francisco José Tapia (fjtapia@gmail.com )\n
///         Distributed under the Boost Software License, Version 1.0.\n
///         ( See accompanyingfile LICENSE_1_0.txt or copy at
///           http://www.boost.org/LICENSE_1_0.txt  )
/// @version 0.1
///
/// @remarks
//-----------------------------------------------------------------------------
#ifndef __CNTREE_SORT_MERGE_SORT_HPP
#define __CNTREE_SORT_MERGE_SORT_HPP

#include <functional>
#include <memory>
#include <type_traits>
#include <vector>
#include <boost/countertree/util/util_iterator.hpp>
#include <boost/countertree/util/sort/insertion_sort.hpp>
#include <boost/countertree/util/sort/token.hpp>
#include <boost/countertree/util/sort/indirect_sort.hpp>



namespace boost
{
namespace countertree
{
namespace util
{
namespace sort
{

template < class iter_t,
           typename compare=std::less<typename iter_value<iter_t>::type >
         >
struct merge_sort_tag
{
//------------------------------------------------------------------------
//                 Definitions and constants
//------------------------------------------------------------------------
typedef typename iter_value<iter_t>::type           value_t ;
static const uint32_t MSort_min = 36 ;

//---------------------------------------------------------------------------
//  function : merge_sort
/// @brief Merge Sort algorithm ( stable sort)
/// @param [in] first : first iterator to the elements to sort
/// @param [in] last : last iterator to the elements to sort
/// @param [in] P : pointer of a data area with at least (N+1)/2 elements
/// @exception
/// @return
/// @remarks
//---------------------------------------------------------------------------
merge_sort_tag (iter_t first, iter_t last, compare comp = compare() )
{   //-------------------------- begin ------------------------------
    auto N = last - first ;
    assert ( N >= 0 );
    if ( N < MSort_min)
    {   insertion_sort ( first, last , comp);
        return  ;
    };
    auto N2 = ( N+1)>>1  ;
    value_t * Ptr = std::get_temporary_buffer<value_t> ( N2 ).first;
    if ( Ptr == nullptr) throw std::bad_alloc ( );
    merge_sort_internal ( first , last , comp,Ptr );
    std::return_temporary_buffer ( Ptr) ;
};

//---------------------------------------------------------------------------
//  function : merge_sort_tag
/// @brief Merge Sort algorithm ( stable sort)
/// @param [in] first : first iterator to the elements to sort
/// @param [in] last : last iterator to the elements to sort
/// @param [in] P : pointer of a data area with at least (N+1)/2 elements
/// @exception
/// @return
/// @remarks
//---------------------------------------------------------------------------
static void merge_sort_internal ( iter_t first , iter_t last,
                                  compare comp, value_t *P )
{   //---------------------------- begin ------------------------------------
    auto N = last - first ;
    assert ( N >= 0 );
    size_t N2 = ( N+1)>>1 ;

    if ( N2 > MSort_min)
    {   merge_sort_internal ( first , first + N2, comp,P );
        merge_sort_internal ( first + N2 , last , comp, P );
    }
    else
    {   insertion_sort ( first , first + N2,comp );
        insertion_sort ( first + N2 , last ,comp );
    };
    //-----------------------------------------------------------------------
    // copy the first N2 elements in P
    //-----------------------------------------------------------------------
    iter_t src = first;
    value_t *dest = P ;
    for ( size_t i =0 ; i < N2 ; ++i)
         ::new ( static_cast<void*>((dest++))) value_t( std::move(*(src++)) );

    //-----------------------------------------------------------------------
    // Fusion
    //-----------------------------------------------------------------------
    iter_t P2 = first + N2 , P2_end= last, Alfa = first ;;
    value_t *P1 = P , *P1_end = P + N2 ;
    bool SW ;
    for ( ;P2 != P2_end and Alfa != P2; ++Alfa   )
    {   //*(Alfa++)= (comp(*P2,*P1))?std::move(*(P2++)):std::move(*(P1++));
        SW = comp(*P2,*P1);
        *Alfa= (SW )?std::move(*(P2++)):std::move(*P1);
        if ( not SW ) (P1++)->~value_t() ;
    }
    if ( P2 == P2_end)
    {   for (; P1!= P1_end; ++P1 )
        {   *(Alfa++) =std::move(*P1);
            P1->~value_t() ;
        };
    };
    //for ( P1 = P ; P1!= P1_end ; ++P1) P1->~value_t() ;
//----------------------------------------------------------------------------
};//        End function merge_sort_tag
//----------------------------------------------------------------------------
//
//----------------------------------------------------------------------------
};//        End of class merge_sort_tag
//----------------------------------------------------------------------------

//---------------------------------------------------------------------------
//  function : merge_sort
/// @brief Merge Sort algorithm ( stable sort)
/// @param [in] first : first iterator to the elements to sort
/// @param [in] last : last iterator to the elements to sort
/// @param [in] P : pointer of a data area with at least (N+1)/2 elements
/// @exception
/// @return
/// @remarks
//---------------------------------------------------------------------------
template < class iter_t,
           typename compare = std::less<typename iter_value<iter_t>::type>
         >
void merge_sort (iter_t first, iter_t last, compare comp = compare() )
{   //----------------------------- begin ----------------------------
    merge_sort_tag<iter_t,compare> ( first, last , comp);
};
template < class iter_t,
           typename compare = std::less<typename iter_value<iter_t>::type>
         >
void stable_sort (iter_t first, iter_t last, compare comp = compare() )
{   //----------------------------- begin ----------------------------
    merge_sort_tag<iter_t,compare> ( first, last , comp);
};

template < class iter_t,
           typename compare = std::less <typename iter_value<iter_t>::type>
         >
void indirect_merge_sort ( iter_t first, iter_t last,compare comp = compare())
{
    indirect_sort_method<merge_sort_tag,iter_t,compare> (first,last,compare());
};
template < class iter_t,
           typename compare = std::less <typename iter_value<iter_t>::type>
         >
void indirect_stable_sort ( iter_t first, iter_t last,compare comp = compare())
{
    indirect_sort_method<merge_sort_tag, iter_t,compare> (first,last,compare());
};
//
//****************************************************************************
};//    End namespace sort
};//    End namespace util
};//    End namespace countertree
};//    End namespace boost
//****************************************************************************
//
#endif
