//----------------------------------------------------------------------------
/// @file selection_sort.hpp
/// @brief Selection Sort algorithm
///
/// @author Copyright (c) 2010 Francisco José Tapia (fjtapia@gmail.com )\n
///         Distributed under the Boost Software License, Version 1.0.\n
///         ( See accompanyingfile LICENSE_1_0.txt or copy at
///           http://www.boost.org/LICENSE_1_0.txt  )
/// @version 0.1
///
/// @remarks
//-----------------------------------------------------------------------------
#ifndef __CNTREE_SORT_SELECTION_SORT_HPP
#define __CNTREE_SORT_SELECTION_SORT_HPP

#include <stdexcept>
#include <cstdint>
#include <cassert>
#include <utility> // para std::swap
#include <type_traits>
#include <functional>
#include <boost/countertree/util/util_iterator.hpp>
#include <boost/countertree/util/definition.hpp>
#include <boost/countertree/util/sort/token.hpp>


namespace boost
{
namespace countertree
{
namespace util
{
namespace sort
{
/*
//-----------------------------------------------------------------------------
//  function : selection_sort
/// @brief : Selection sort algorithm
/// @param [in] first: iterator to the first element of the range
/// @param [in] last : iterator to the next element of the last in the range
/// @remarks This algorithm is O(N²)
//-----------------------------------------------------------------------------
template< class iter_t >
inline void selection_sort( iter_t first, iter_t last ) NOEXCEPT
{   //------------------------- begin ----------------------
    if ( (last-first) < 2) return;

    NOALIAS iter_t Aux1 = first, Aux2 = first ;
    for ( NOALIAS iter_t Alfa = first +1 ; Alfa != last ; ++Alfa)
    {   Aux1 = Aux2 = Alfa -1 ;
        for ( NOALIAS iter_t Beta = Alfa ; Beta != last ; ++Beta )
            if ( *Beta < *Aux1) Aux1 = Beta ;
        if ( Aux1 != Aux2 ) std::swap ( *Aux1 , *Aux2);
    };
};
*/
//-----------------------------------------------------------------------------
//  function : selection_sort
/// @brief : Selection sort algorithm
/// @param [in] first: iterator to the first element of the range
/// @param [in] last : iterator to the next element of the last in the range
/// @param [in] comp : object for to do the comparison between the elements
/// @remarks This algorithm is O(N²)
//-----------------------------------------------------------------------------
template < class iter_t,
           typename compare = std::less<typename iter_value<iter_t>::type>
         >
inline void selection_sort( iter_t first, iter_t last, compare comp = compare()) NOEXCEPT
{   //------------------------- begin ----------------------
    if ( (last-first) <2) return;

    NOALIAS iter_t Aux1 = first, Aux2 = first ;
    for ( NOALIAS iter_t Alfa = first +1 ; Alfa != last ; ++Alfa)
    {   Aux1 = Aux2 = Alfa -1 ;
        for ( NOALIAS iter_t Beta = Alfa ; Beta != last ; ++Beta )
            if ( comp (*Beta,*Aux1) )
                Aux1 = Beta ;
        if ( Aux1 != Aux2 ) std::swap ( *Aux1 , *Aux2);
    };
};
template < class iter_t,
           typename compare = std::less<typename iter_value<iter_t>::type>
         >
struct selection_tag
{   selection_tag( iter_t first, iter_t last, compare comp = compare()) NOEXCEPT
    {   selection_sort ( first,last,comp);
    };
};


/*
template < class iter_t,
           typename compare = std::less<typename iter_value<iter_t>::type>
         >
struct selection_sort
{
selection_sort( iter_t first, iter_t last, compare comp = compare()) NOEXCEPT

{   //------------------------- begin ----------------------
    if ( (last-first) <2) return;

    NOALIAS iter_t Aux1 = first, Aux2 = first ;
    for ( NOALIAS iter_t Alfa = first +1 ; Alfa != last ; ++Alfa)
    {   Aux1 = Aux2 = Alfa -1 ;
        for ( NOALIAS iter_t Beta = Alfa ; Beta != last ; ++Beta )
            if ( comp (*Beta,*Aux1)) Aux1 = Beta ;
        if ( Aux1 != Aux2 ) std::swap ( *Aux1 , *Aux2);
    };
};
};
*/
//
//****************************************************************************
};//    End namespace sort
};//    End namespace util
};//    End namespace countertree
};//    End namespace boost
//****************************************************************************
//
#endif
